from typing import Iterable
from typing import Optional
from typing import Union

from .request import Request
from .response import Response


class CORSMiddleware(object):
    """CORS Middleware.

    This middleware provides a simple out-of-the box CORS policy, including handling
    of preflighted requests from the browser.

    See also:

    * https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
    * https://www.w3.org/TR/cors/#resource-processing-model

    Keyword Arguments:
        allow_origins (Union[str, Iterable[str]]): List of origins to allow (case
            sensitive). The string ``'*'`` acts as a wildcard, matching every origin.
            (default ``'*'``).
        expose_headers (Optional[Union[str, Iterable[str]]]): List of additional response
            headers to expose via the ``Access-Control-Expose-Headers`` header.
            These headers are in addition to the CORS-safelisted ones:
            ``Cache-Control``, ``Content-Language``, ``Content-Length``, ``Content-Type``,
            ``Expires``, ``Last-Modified``, ``Pragma``. (default ``None``).

            See also:
            https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers
        allow_credentials (Optional[Union[str, Iterable[str]]]): List of origins (case sensitive)
            for which to allow credentials via the ``Access-Control-Allow-Credentials`` header.
            The string ``'*'`` acts as a wildcard, matching every allowed origin, while
            ``None`` disallows all origins. This parameter takes effect only
            if the origin is allowed by the ``allow_origins`` argument. (Default ``None``).

    """
    def __init__(
        self,
        allow_origins: Union[str, Iterable[str]] = '*',
        expose_headers: Optional[Union[str, Iterable[str]]] = None,
        allow_credentials: Optional[Union[str, Iterable[str]]] = None,
    ):
        if allow_origins == '*':
            self.allow_origins = allow_origins
        else:
            if isinstance(allow_origins, str):
                allow_origins = [allow_origins]
            self.allow_origins = frozenset(allow_origins)
            if '*' in self.allow_origins:
                raise ValueError(
                    'The wildcard string "*" may only be passed to allow_origins as a '
                    'string literal, not inside an iterable.'
                )

        if expose_headers is not None and not isinstance(expose_headers, str):
            expose_headers = ', '.join(expose_headers)
        self.expose_headers = expose_headers

        if allow_credentials is None:
            allow_credentials = frozenset()
        elif allow_credentials != '*':
            if isinstance(allow_credentials, str):
                allow_credentials = [allow_credentials]
            allow_credentials = frozenset(allow_credentials)
            if '*' in allow_credentials:
                raise ValueError(
                    'The wildcard string "*" may only be passed to allow_credentials as a '
                    'string literal, not inside an iterable.'
                )
        self.allow_credentials = allow_credentials

    def process_response(self, req: Request, resp: Response, resource, req_succeeded):
        """Implement the CORS policy for all routes.

        This middleware provides a simple out-of-the box CORS policy,
        including handling of preflighted requests from the browser.

        See also: https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

        See also: https://www.w3.org/TR/cors/#resource-processing-model
        """

        origin = req.get_header('Origin')
        if origin is None:
            return

        if self.allow_origins != '*' and origin not in self.allow_origins:
            return

        if resp.get_header('Access-Control-Allow-Origin') is None:
            set_origin = '*' if self.allow_origins == '*' else origin
            if self.allow_credentials == '*' or origin in self.allow_credentials:
                set_origin = origin
                resp.set_header('Access-Control-Allow-Credentials', 'true')
            resp.set_header('Access-Control-Allow-Origin', set_origin)

        if self.expose_headers:
            resp.set_header('Access-Control-Expose-Headers', self.expose_headers)

        if (req_succeeded and
                req.method == 'OPTIONS' and
                req.get_header('Access-Control-Request-Method')):

            # NOTE(kgriffs): This is a CORS preflight request. Patch the
            #   response accordingly.

            allow = resp.get_header('Allow')
            resp.delete_header('Allow')

            allow_headers = req.get_header('Access-Control-Request-Headers', default='*')

            resp.set_header('Access-Control-Allow-Methods', allow)
            resp.set_header('Access-Control-Allow-Headers', allow_headers)
            resp.set_header('Access-Control-Max-Age', '86400')  # 24 hours

    async def process_response_async(self, *args):
        self.process_response(*args)



import hashlib
import os
import random
import subprocess
import sys
import time

import pytest
import requests

from falcon import testing

_HERE = os.path.abspath(os.path.dirname(__file__))
_SERVER_HOST = '127.0.0.1'
_SIZE_1_KB = 1024
_SIZE_1_MB = _SIZE_1_KB ** 2

_REQUEST_TIMEOUT = 10
_STARTUP_TIMEOUT = 10
_SHUTDOWN_TIMEOUT = 20


def _gunicorn_args(host, port, extra_opts=()):
    """Gunicorn"""
    try:
        import gunicorn  # noqa: F401
    except ImportError:
        pytest.skip('gunicorn not installed')

    args = (
        'gunicorn',
        '--access-logfile', '-',
        '--bind', '{}:{}'.format(host, port),

        # NOTE(vytas): Although rare, but Meinheld workers have been noticed to
        #   occasionally hang on shutdown.
        '--graceful-timeout', str(_SHUTDOWN_TIMEOUT // 2),
        # NOTE(vytas): In case a worker hangs for an unexpectedly long time
        #   while reading or processing request (the default value is 30).
        '--timeout', str(_REQUEST_TIMEOUT),
    )
    return args + extra_opts + ('_wsgi_test_app:app',)


def _meinheld_args(host, port):
    """Gunicorn + Meinheld"""
    try:
        import meinheld  # noqa: F401
    except ImportError:
        pytest.skip('meinheld not installed')

    return _gunicorn_args(
        host,
        port,
        (
            '--workers', '2',
            '--worker-class', 'egg:meinheld#gunicorn_worker',
        ),
    )


def _uvicorn_args(host, port):
    """Uvicorn (WSGI interface)"""
    try:
        import uvicorn  # noqa: F401
    except ImportError:
        pytest.skip('uvicorn not installed')

    return (
        'uvicorn',
        '--host', host,
        '--port', str(port),
        '--interface', 'wsgi',
        '_wsgi_test_app:app',
    )


def _uwsgi_args(host, port):
    """uWSGI"""
    return (
        'uwsgi',
        '--http', '{}:{}'.format(host, port),
        '--wsgi-file', '_wsgi_test_app.py',
    )


def _waitress_args(host, port):
    """Waitress"""
    try:
        import waitress  # noqa: F401
    except ImportError:
        pytest.skip('waitress not installed')

    return (
        'waitress-serve',
        '--listen', '{}:{}'.format(host, port),
        '_wsgi_test_app:app',
    )


@pytest.fixture(params=[
    _gunicorn_args,
    _meinheld_args,
    _uvicorn_args,
    _uwsgi_args,
    _waitress_args,
])
def server_url(request):
    if sys.platform.startswith('win'):
        pytest.skip('WSGI server tests are currently unsupported on Windows')

    server_args = request.param

    for attempt in range(3):
        server_port = testing.get_unused_port()
        base_url = 'http://{}:{}'.format(_SERVER_HOST, server_port)

        args = server_args(_SERVER_HOST, server_port)
        print('Starting {}...'.format(server_args.__doc__))
        print(' '.join(args))
        try:
            server = subprocess.Popen(args, cwd=_HERE)
        except FileNotFoundError:
            pytest.skip('{} executable is not installed'.format(args[0]))

        # NOTE(vytas): give the app server some time to start.
        start_time = time.time()
        while time.time() - start_time < _STARTUP_TIMEOUT:
            try:
                requests.get(base_url + '/hello', timeout=0.2)
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
                time.sleep(0.2)
            else:
                break
        else:
            if server.poll() is None:
                pytest.fail('Server is not responding to requests')
            else:
                # NOTE(kgriffs): The server did not start up; probably due to
                #   the port being in use. We could check the output but
                #   capsys fixture may not have buffered the error output
                #   yet, so we just retry.
                continue

        yield base_url
        break

    else:
        pytest.fail('Could not start server')

    print('\n[Sending SIGTERM to server process...]')
    server.terminate()

    try:
        server.communicate(timeout=_SHUTDOWN_TIMEOUT)
    except subprocess.TimeoutExpired:
        print('\n[Killing stubborn server process...]')

        server.kill()
        server.communicate()

        pytest.fail('Server process did not exit in a timely manner and had to be killed.')


class TestWSGIServer:

    def test_get(self, server_url):
        resp = requests.get(server_url + '/hello', timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text == 'Hello, World!\n'
        assert resp.headers.get('Content-Type') == 'text/plain; charset=utf-8'
        assert resp.headers.get('X-Falcon') == 'peregrine'

    def test_get_deprecated(self, server_url):
        resp = requests.get(server_url + '/deprecated', timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text == 'Hello, World!\n'
        assert resp.headers.get('Content-Type') == 'text/plain; charset=utf-8'
        assert resp.headers.get('X-Falcon') == 'deprecated'

    def test_post_multipart_form(self, server_url):
        size = random.randint(8 * _SIZE_1_MB, 15 * _SIZE_1_MB)
        data = os.urandom(size)
        digest = hashlib.sha1(data).hexdigest()
        files = {
            'random': ('random.dat', data),
            'message': ('hello.txt', b'Hello, World!\n'),
        }

        resp = requests.post(
            server_url + '/forms', files=files, timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.json() == {
            'message': {
                'filename': 'hello.txt',
                'sha1': '60fde9c2310b0d4cad4dab8d126b04387efba289',
            },
            'random': {
                'filename': 'random.dat',
                'sha1': digest,
            },
        }

    def test_static_file(self, server_url):
        resp = requests.get(server_url + '/tests/test_wsgi_servers.py', timeout=_REQUEST_TIMEOUT)
        assert resp.status_code == 200
        assert resp.text.startswith(
            'import hashlib\n'
            'import os\n'
            'import random\n'
            'import subprocess\n'
            'import sys\n'
            'import time\n'
        )
        assert resp.headers.get('Content-Disposition') == (
            'attachment; filename="test_wsgi_servers.py"')



from functools import partial
import json

from falcon import errors
from falcon import http_error
from falcon.media.base import BaseHandler
from falcon.media.base import TextBaseHandlerWS


class JSONHandler(BaseHandler):
    """JSON media handler.

    This handler uses Python's standard :py:mod:`json` library by default, but
    can be easily configured to use any of a number of third-party JSON
    libraries, depending on your needs. For example, you can often
    realize a significant performance boost under CPython by using an
    alternative library. Good options in this respect include `orjson`,
    `python-rapidjson`, and `mujson`.

    This handler will raise a :class:`falcon.MediaNotFoundError` when attempting
    to parse an empty body, or a :class:`falcon.MediaMalformedError`
    if an error happens while parsing the body.

    Note:
        If you are deploying to PyPy, we recommend sticking with the standard
        library's JSON implementation, since it will be faster in most cases
        as compared to a third-party library.

    Overriding the default JSON implementation is simply a matter of specifying
    the desired ``dumps`` and ``loads`` functions::

        import falcon
        from falcon import media

        import rapidjson

        json_handler = media.JSONHandler(
            dumps=rapidjson.dumps,
            loads=rapidjson.loads,
        )
        extra_handlers = {
            'application/json': json_handler,
        }

        app = falcon.App()
        app.req_options.media_handlers.update(extra_handlers)
        app.resp_options.media_handlers.update(extra_handlers)

    By default, ``ensure_ascii`` is passed to the ``json.dumps`` function.
    If you override the ``dumps`` function, you will need to explicitly set
    ``ensure_ascii`` to ``False`` in order to enable the serialization of
    Unicode characters to UTF-8. This is easily done by using
    :any:`functools.partial` to apply the desired keyword argument. In fact, you
    can use this same technique to customize any option supported by the
    ``dumps`` and ``loads`` functions::

        from functools import partial

        from falcon import media
        import rapidjson

        json_handler = media.JSONHandler(
            dumps=partial(
                rapidjson.dumps,
                ensure_ascii=False, sort_keys=True
            ),
        )

    Keyword Arguments:
        dumps (func): Function to use when serializing JSON responses.
        loads (func): Function to use when deserializing JSON requests.
    """

    def __init__(self, dumps=None, loads=None):
        self._dumps = dumps or partial(json.dumps, ensure_ascii=False)
        self._loads = loads or json.loads

        # PERF(kgriffs): Test dumps once up front so we can set the
        #     proper serialize implementation.
        result = self._dumps({'message': 'Hello World'})
        if isinstance(result, str):
            self.serialize = self._serialize_s
            self.serialize_async = self._serialize_async_s
        else:
            self.serialize = self._serialize_b
            self.serialize_async = self._serialize_async_b

        # NOTE(kgriffs): To be safe, only enable the optimized protocol when
        #   not subclassed.
        if type(self) is JSONHandler:
            self._serialize_sync = self.serialize
            self._deserialize_sync = self._deserialize

    def _deserialize(self, data):
        if not data:
            raise errors.MediaNotFoundError('JSON')
        try:
            return self._loads(data.decode())
        except ValueError as err:
            raise errors.MediaMalformedError('JSON') from err

    def deserialize(self, stream, content_type, content_length):
        return self._deserialize(stream.read())

    async def deserialize_async(self, stream, content_type, content_length):
        return self._deserialize(await stream.read())

    # NOTE(kgriffs): Make content_type a kwarg to support the
    #   Request.render_body() shortcut optimization.
    def _serialize_s(self, media, content_type=None) -> bytes:
        return self._dumps(media).encode()

    async def _serialize_async_s(self, media, content_type) -> bytes:
        return self._dumps(media).encode()

    def _serialize_b(self, media, content_type) -> bytes:
        return self._dumps(media)

    async def _serialize_async_b(self, media, content_type) -> bytes:
        return self._dumps(media)


class JSONHandlerWS(TextBaseHandlerWS):
    """WebSocket media handler for de(serializing) JSON to/from TEXT payloads.

    This handler uses Python's standard :py:mod:`json` library by default, but
    can be easily configured to use any of a number of third-party JSON
    libraries, depending on your needs. For example, you can often
    realize a significant performance boost under CPython by using an
    alternative library. Good options in this respect include `orjson`,
    `python-rapidjson`, and `mujson`.

    Note:
        If you are deploying to PyPy, we recommend sticking with the standard
        library's JSON implementation, since it will be faster in most cases
        as compared to a third-party library.

    Overriding the default JSON implementation is simply a matter of specifying
    the desired ``dumps`` and ``loads`` functions::

        import falcon
        from falcon import media

        import rapidjson

        json_handler = media.JSONHandlerWS(
            dumps=rapidjson.dumps,
            loads=rapidjson.loads,
        )

        app = falcon.asgi.App()
        app.ws_options.media_handlers[falcon.WebSocketPayloadType.TEXT] = json_handler

    By default, ``ensure_ascii`` is passed to the ``json.dumps`` function.
    If you override the ``dumps`` function, you will need to explicitly set
    ``ensure_ascii`` to ``False`` in order to enable the serialization of
    Unicode characters to UTF-8. This is easily done by using
    :any:`functools.partial` to apply the desired keyword argument. In fact, you
    can use this same technique to customize any option supported by the
    ``dumps`` and ``loads`` functions::

        from functools import partial

        from falcon import media
        import rapidjson

        json_handler = media.JSONHandlerWS(
            dumps=partial(
                rapidjson.dumps,
                ensure_ascii=False, sort_keys=True
            ),
        )

    Keyword Arguments:
        dumps (func): Function to use when serializing JSON.
        loads (func): Function to use when deserializing JSON.
    """

    __slots__ = ['dumps', 'loads']

    def __init__(self, dumps=None, loads=None):
        self._dumps = dumps or partial(json.dumps, ensure_ascii=False)
        self._loads = loads or json.loads

    def serialize(self, media: object) -> str:
        return self._dumps(media)

    def deserialize(self, payload: str) -> object:
        return self._loads(payload)


http_error._DEFAULT_JSON_HANDLER = _DEFAULT_JSON_HANDLER = JSONHandler()  # type: ignore



import pytest

import falcon
from falcon import testing
from falcon.asgi import App


@pytest.mark.asyncio
async def test_default_headers():
    class Resource:
        async def on_get(self, req, resp):
            resp.set_header('the-answer', req.get_header('the-answer'))

    app = App()
    app.add_route('/', Resource())
    client = testing.TestClient(app, headers={'the-answer': '42'})

    async with client as conductor:
        result = await conductor.simulate_get()
        assert result.headers['the-answer'] == '42'


@pytest.mark.asyncio
async def test_wsgi_not_supported():
    with pytest.raises(falcon.CompatibilityError):
        async with testing.TestClient(falcon.App()):
            pass

    with pytest.raises(falcon.CompatibilityError):
        async with testing.ASGIConductor(falcon.App()):
            pass


@pytest.mark.parametrize('method', ['get', 'head', 'post', 'put', 'options', 'patch', 'delete'])
@pytest.mark.asyncio
async def test_responders(method):
    class Resource:
        async def on_get(self, req, resp):
            resp.set_header('method', 'get')

        async def on_head(self, req, resp):
            resp.set_header('method', 'head')

        async def on_post(self, req, resp):
            resp.set_header('method', 'post')

        async def on_put(self, req, resp):
            resp.set_header('method', 'put')

        async def on_options(self, req, resp):
            resp.set_header('method', 'options')

        async def on_patch(self, req, resp):
            resp.set_header('method', 'patch')

        async def on_delete(self, req, resp):
            resp.set_header('method', 'delete')

    resource = Resource()

    app = App()
    app.add_route('/', resource)

    async with testing.ASGIConductor(app) as conductor:
        simulate = getattr(conductor, 'simulate_' + method)
        result = await simulate('/')
        assert result.headers['method'] == method



#!/usr/bin/env python

import atexit
import pathlib
import subprocess

import toml

HERE = pathlib.Path(__file__).resolve().parent
ROOT = HERE.parent


def _write_changelog(target, data):
    with open(ROOT / target, 'wb') as rst:
        rst.write(data)


def get_target_filename():
    with open(ROOT / 'pyproject.toml') as pyproject_toml:
        project = toml.load(pyproject_toml)
    return project['tool']['towncrier']['filename']


def render_draft(target):
    with open(ROOT / target, 'rb') as rst:
        template = rst.read()
    # NOTE(vytas): Restore the template once we are done.
    atexit.register(_write_changelog, target, template)

    draft = subprocess.check_output(('towncrier', '--draft'), cwd=ROOT)
    # NOTE(vytas): towncrier --draft does not seem to use the template,
    #   so we substitute manually.
    rendered = template.replace(b'.. towncrier release notes start', draft, 1)
    _write_changelog(target, rendered)


def build_docs():
    subprocess.check_call(
        (
            'sphinx-build',
            '-W',
            '-E',
            '-b',
            'html',
            ROOT / 'docs',
            ROOT / 'docs/_build/html',
        )
    )


if __name__ == '__main__':
    target = get_target_filename()
    render_draft(target)
    build_docs()



import pytest

import falcon
import falcon.errors as errors
import falcon.status_codes as status
from falcon.util.deprecation import DeprecatedWarning


@pytest.mark.parametrize('err, title', [
    (falcon.HTTPBadRequest, status.HTTP_400),
    (falcon.HTTPUnauthorized, status.HTTP_401),
    (falcon.HTTPForbidden, status.HTTP_403),
    (falcon.HTTPNotFound, status.HTTP_404),
    (errors.HTTPRouteNotFound, status.HTTP_404),
    (falcon.HTTPNotAcceptable, status.HTTP_406),
    (falcon.HTTPConflict, status.HTTP_409),
    (falcon.HTTPGone, status.HTTP_410),
    (falcon.HTTPLengthRequired, status.HTTP_411),
    (falcon.HTTPPreconditionFailed, status.HTTP_412),
    (falcon.HTTPPayloadTooLarge, status.HTTP_413),
    (falcon.HTTPUriTooLong, status.HTTP_414),
    (falcon.HTTPUnsupportedMediaType, status.HTTP_415),
    (falcon.HTTPUnprocessableEntity, status.HTTP_422),
    (falcon.HTTPLocked, status.HTTP_423),
    (falcon.HTTPFailedDependency, status.HTTP_424),
    (falcon.HTTPPreconditionRequired, status.HTTP_428),
    (falcon.HTTPTooManyRequests, status.HTTP_429),
    (falcon.HTTPRequestHeaderFieldsTooLarge, status.HTTP_431),
    (falcon.HTTPUnavailableForLegalReasons, status.HTTP_451),
    (falcon.HTTPInternalServerError, status.HTTP_500),
    (falcon.HTTPNotImplemented, status.HTTP_501),
    (falcon.HTTPBadGateway, status.HTTP_502),
    (falcon.HTTPServiceUnavailable, status.HTTP_503),
    (falcon.HTTPGatewayTimeout, status.HTTP_504),
    (falcon.HTTPVersionNotSupported, status.HTTP_505),
    (falcon.HTTPInsufficientStorage, status.HTTP_507),
    (falcon.HTTPLoopDetected, status.HTTP_508),
    (falcon.HTTPNetworkAuthenticationRequired, status.HTTP_511),
])
def test_with_default_title_and_desc(err, title):
    with pytest.raises(err) as e:
        raise err()

    assert e.value.title == title
    assert e.value.description is None

    if e.value.headers:
        assert 'Retry-After' not in e.value.headers


@pytest.mark.parametrize('err, title, args', (
    (falcon.HTTPMethodNotAllowed, status.HTTP_405, (['GET'], )),
    (falcon.HTTPRangeNotSatisfiable, status.HTTP_416, (11,)),
))
def test_with_default_title_and_desc_args(err, title, args):
    with pytest.raises(err) as e:
        raise err(*args)

    assert e.value.title == title
    assert e.value.description is None

    if e.value.headers:
        assert 'Retry-After' not in e.value.headers


@pytest.mark.parametrize('err', [
    falcon.HTTPBadRequest,
    falcon.HTTPUnauthorized,
    falcon.HTTPForbidden,
    falcon.HTTPNotFound,
    errors.HTTPRouteNotFound,
    falcon.HTTPNotAcceptable,
    falcon.HTTPConflict,
    falcon.HTTPGone,
    falcon.HTTPLengthRequired,
    falcon.HTTPPreconditionFailed,
    falcon.HTTPPayloadTooLarge,
    falcon.HTTPUriTooLong,
    falcon.HTTPUnsupportedMediaType,
    falcon.HTTPUnprocessableEntity,
    falcon.HTTPLocked,
    falcon.HTTPFailedDependency,
    falcon.HTTPPreconditionRequired,
    falcon.HTTPTooManyRequests,
    falcon.HTTPRequestHeaderFieldsTooLarge,
    falcon.HTTPUnavailableForLegalReasons,
    falcon.HTTPInternalServerError,
    falcon.HTTPNotImplemented,
    falcon.HTTPBadGateway,
    falcon.HTTPServiceUnavailable,
    falcon.HTTPGatewayTimeout,
    falcon.HTTPVersionNotSupported,
    falcon.HTTPInsufficientStorage,
    falcon.HTTPLoopDetected,
    falcon.HTTPNetworkAuthenticationRequired,
])
def test_with_title_desc_and_headers(err):
    title = 'trace'
    desc = 'boom'
    headers = {'foo': 'bar'}

    with pytest.raises(err) as e:
        raise err(title=title, description=desc, headers=headers)

    assert e.value.title == title
    assert e.value.description == desc
    assert e.value.headers['foo'] == 'bar'


@pytest.mark.parametrize('err', [
    falcon.HTTPBadRequest,
    falcon.HTTPUnauthorized,
    falcon.HTTPForbidden,
    falcon.HTTPNotFound,
    errors.HTTPRouteNotFound,
    falcon.HTTPNotAcceptable,
    falcon.HTTPConflict,
    falcon.HTTPGone,
    falcon.HTTPLengthRequired,
    falcon.HTTPPreconditionFailed,
    falcon.HTTPPayloadTooLarge,
    falcon.HTTPUriTooLong,
    falcon.HTTPUnsupportedMediaType,
    falcon.HTTPUnprocessableEntity,
    falcon.HTTPLocked,
    falcon.HTTPFailedDependency,
    falcon.HTTPPreconditionRequired,
    falcon.HTTPTooManyRequests,
    falcon.HTTPRequestHeaderFieldsTooLarge,
    falcon.HTTPUnavailableForLegalReasons,
    falcon.HTTPInternalServerError,
    falcon.HTTPNotImplemented,
    falcon.HTTPBadGateway,
    falcon.HTTPServiceUnavailable,
    falcon.HTTPGatewayTimeout,
    falcon.HTTPVersionNotSupported,
    falcon.HTTPInsufficientStorage,
    falcon.HTTPLoopDetected,
    falcon.HTTPNetworkAuthenticationRequired,
])
def test_kw_only(err):
    # only deprecated for now
    # with pytest.raises(TypeError, match='positional argument'):
    #     err('foo', 'bar')
    with pytest.warns(DeprecatedWarning, match='positional args are deprecated'):
        err('foo', 'bar')


@pytest.mark.parametrize('err, args', (
    (falcon.HTTPMethodNotAllowed, (['GET'], )),
    (falcon.HTTPRangeNotSatisfiable, (11,)),
))
def test_with_title_desc_and_headers_args(err, args):
    title = 'trace'
    desc = 'boom'
    headers = {'foo': 'bar'}

    with pytest.raises(err) as e:
        raise err(*args, title=title, description=desc, headers=headers)

    assert e.value.title == title
    assert e.value.description == desc
    assert e.value.headers['foo'] == 'bar'


@pytest.mark.parametrize('err, args', (
    (falcon.HTTPMethodNotAllowed, (['GET'], )),
    (falcon.HTTPRangeNotSatisfiable, (11,)),
    (falcon.HTTPInvalidHeader, ('foo', 'bar')),
    (falcon.HTTPMissingHeader, ('foo',)),
    (falcon.HTTPInvalidParam, ('foo', 'bar')),
    (falcon.HTTPMissingParam, ('foo',)),
))
def test_args_kw_only(err, args):
    # only deprecated for now
    # with pytest.raises(TypeError, match='positional argument'):
    #     err(*args, 'bar')
    with pytest.warns(DeprecatedWarning, match='positional args are deprecated'):
        err(*args, 'bar')


@pytest.mark.parametrize('err', [
    falcon.HTTPServiceUnavailable,
    falcon.HTTPTooManyRequests,
    falcon.HTTPPayloadTooLarge,
])
def test_with_retry_after(err):
    with pytest.raises(err) as e:
        raise err(retry_after='123')

    assert e.value.headers['Retry-After'] == '123'


@pytest.mark.parametrize('err', [
    falcon.HTTPServiceUnavailable,
    falcon.HTTPTooManyRequests,
    falcon.HTTPPayloadTooLarge,
])
def test_with_retry_after_and_headers(err):
    with pytest.raises(err) as e:
        raise err(retry_after='123', headers={'foo': 'bar'})

    assert e.value.headers['Retry-After'] == '123'
    assert e.value.headers['foo'] == 'bar'


def test_http_error_repr():
    error = falcon.HTTPBadRequest()
    _repr = '<%s: %s>' % (error.__class__.__name__, error.status)
    assert error.__repr__() == _repr


@pytest.mark.parametrize('err, args, title, desc', (
    (falcon.HTTPInvalidHeader, ('foo', 'bar'), 'Invalid header value',
     'The value provided for the "bar" header is invalid. foo'),
    (falcon.HTTPMissingHeader, ('foo',), 'Missing header value', 'The "foo" header is required.'),
    (falcon.HTTPInvalidParam, ('foo', 'bar'), 'Invalid parameter',
     'The "bar" parameter is invalid. foo'),
    (falcon.HTTPMissingParam, ('foo',), 'Missing parameter', 'The "foo" parameter is required.'),
))
def test_custom_400(err, args, title, desc):
    with pytest.raises(err) as e:
        raise err(*args)

    assert e.value.title == title
    assert e.value.description == desc


@pytest.mark.parametrize('err, header_name, kw_name, args, res, kw_required', (
    (falcon.HTTPUnauthorized, 'WWW-Authenticate', 'challenges', ('a', 'b'), 'a, b', False),
    (falcon.HTTPMethodNotAllowed, 'Allow', 'allowed_methods', ('a', 'b'), 'a, b', True),
    (falcon.HTTPPayloadTooLarge, 'Retry-After', 'retry_after', 123, '123', False),
    (falcon.HTTPRangeNotSatisfiable, 'Content-Range', 'resource_length', 123, 'bytes */123', True),
    (falcon.HTTPTooManyRequests, 'Retry-After', 'retry_after', 123, '123', False),
    (falcon.HTTPServiceUnavailable, 'Retry-After', 'retry_after', 123, '123', False),
))
class TestErrorsWithHeadersKW:
    def test_no_header(self, err, header_name, kw_name, args, res, kw_required):
        if not kw_required:
            value = err()

            if value.headers:
                assert header_name not in value.headers

    def test_other_header(self, err, header_name, kw_name, args, res, kw_required):
        headers = {'foo bar': 'baz'}
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert value.headers[header_name] == res

    def test_override_header(self, err, header_name, kw_name, args, res, kw_required):
        headers = {'foo bar': 'baz', header_name: 'other'}
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert value.headers[header_name] == res

    def test_other_header_list(self, err, header_name, kw_name, args, res, kw_required):
        headers = [('foo bar', 'baz')]
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert isinstance(value.headers, dict)
        assert value.headers[header_name] == res

    def test_override_header_list(self, err, header_name, kw_name, args, res, kw_required):
        headers = [('foo bar', 'baz'), (header_name, 'other')]
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert isinstance(value.headers, dict)
        assert value.headers[header_name] == res



import asyncio
import time


async def _say_hi():
    print(f'[{time.time()}] Hi!')


async def app(scope, receive, send):
    await send({
        'type': 'http.response.start',
        'status': 200,
        'headers': [
            [b'content-type', b'application/json'],
        ]
    })
    await send({
        'type': 'http.response.body',
        'body': f'[{time.time()}] Hello world!'.encode(),
    })

    loop = asyncio.get_event_loop()
    loop.create_task(_say_hi())
